#include "FSM.h"



namespace FSM_ns {

double	FSM::get_scan_calculated_time(double s, double v, double a, double d){
	double t = -1;

	if ( (s == 0.0) || (v <= 0.0) || (a <= 0.0) || (d <= 0.0) ) return -1;
	double _s = abs(s);

	// tempo in accel per arrivare da 0 a v:
	double ta = v/a;

	// tempo in decl per arrivare da v a 0:
	double td = v/d;

	// spazio percorso in accel:
	double sa = 0.5 * v * v / a;

	// spazio percorso in decel:
	double sd = 0.5 * v * v / d;

	double tv = 0.0;
	double sv = 0.0;
	if ( sa + sd < _s ) {
		tv = (_s - sa - sd)/v;
		t = ta + tv + td;
		sv = v * tv;
	} else {
		td = sqrt(2*_s/(d*d/a + d));
		sd = 0.5 * d * td * td;
		ta = d/a * td;
		sa = 0.5 * a * ta * ta;
		t = ta + td;
	}

	#ifdef _DBG_FSM_CALCULATED_TIME
		cout << "s: " << _s << " " << "v: " << v << " " << "a: " << a << " " << "d: " << d << endl;
		cout << "ta: " << ta << " sa: " << sa << endl;
		cout << "tv: " << tv << " sv: " << sv << endl;
		cout << "td: " << td << " sd: " << sd << endl;
		cout << " ---- t = " << t << " ----" << endl;
	#endif

	
	return t;
}


void FSM::do_stop_1(){
	std::stringstream ss;
	std::string obj("FSM::fsm()");
	ss.str("");
       	ss << "do_stop_1()";
       	prtverbose(1, obj, ss.str());

	AxisG2Scan_ns::sp_ac_dc_t axis_default_data = pdev->get_axis_default_data();

	vector<Tango::DeviceAttribute> attr_in;
	string att1_name("Speed");
	string att2_name("Acceleration");
	string att3_name("Deceleration");

       	#ifdef _DBG_FSM_STOP_1
       		cout << _name << " : axis_default_data.valid : " << (axis_default_data.valid?"VALID":"NOTvalid")  << endl;
       		cout << _name << " : axis_default_data.speed : " << axis_default_data.speed  << endl;
       		cout << _name << " : axis_default_data.accel : " << axis_default_data.accel  << endl;
       		cout << _name << " : axis_default_data.decel : " << axis_default_data.decel  << endl;
       	#endif /* _DBG_FSM_STOP_1 */

	attr_in.push_back(Tango::DeviceAttribute(att1_name, axis_default_data.speed));
	attr_in.push_back(Tango::DeviceAttribute(att2_name, axis_default_data.accel));
	attr_in.push_back(Tango::DeviceAttribute(att3_name, axis_default_data.decel));
	Tango::DeviceAttribute da3(att3_name, axis_default_data.decel);
	Tango::DeviceAttribute da2(att2_name, axis_default_data.accel);
	Tango::DeviceAttribute da1(att1_name, axis_default_data.speed);

	struct timeval t0, t1, dt;

       	string msg = "do_stop_1() 1:: ";
       	try {
		gettimeofday(&t0, NULL);
       		pdev->get_axis_dp()->write_attributes(attr_in);

       		//pdev->get_axis_dp()->write_attribute(da1);
       		//pdev->get_axis_dp()->write_attribute(da2);
       		//pdev->get_axis_dp()->write_attribute(da3);
		gettimeofday(&t1, NULL);

       	} catch(Tango::DevFailed e) {

		gettimeofday(&t1, NULL);
		trace_err = timestamp() + "do_stop_1() 1:: Tango Exception";
       		cout << msg << endl;
                Tango::Except::print_exception(e);
		cout << "---" << endl;
       		_tr     = FSMT_NOP;
       		_FSM  = FSM_STOP_ERROR;
       	} catch (...) {
		trace_err = timestamp() + "do_stop_1() 1:: General Exception";
       		cout << msg << endl;
		cout << "---" << endl;
       		_tr     = FSMT_NOP;
       		_FSM  = FSM_STOP_ERROR;
       	}

	timersub(&t1, &t0, &dt);
       	#ifdef _DBG_FSM_STOP_1
       		cout << _name << " : get_axis_dp()->write_attributes(attr_in) time: : " << (dt.tv_sec + ((float)dt.tv_usec)/1000000)  << endl;
       	#endif /* _DBG_FSM_STOP_1 */
}


FSM_t	FSM::fsm(FSMTransitions_t tr){	// proceed with FSM
        std::stringstream ss;
	std::string obj("FSM::fsm()");


        #ifdef _DBG_TRACE_FSM
                ss.str("");
                ss << "== == == FSM: " << FSMName[_FSM] << " & new FSMT: " << FSMTransitions[tr];
                prtverbose(0, obj, ss.str());
        #endif /* _DBG_TRACE_FSM */


        ss.str("");
        ss << "FSM(" << FSMName[_FSM] << " : new= " << FSMTransitions[tr] << ")";
        prtverbose(2, obj, ss.str());

        if (tr == FSMT_ABORT){
                        _tr    = FSMT_NOP;
                        _FSM      = FSM_WAIT;
                        _FSMerror = FSM_E000;
                        return _FSM;
        }

        //=========================
        // accept tr only in state:

        switch (_FSM){
                case FSM_WAIT:
                        _tr = tr;
                        break;
                default:
        		if (tr == FSMT_STOP) _tr = tr;
                        break;
        }

        //=========================
        // handle transition

        switch (_tr) {

                default:
                case FSMT_NOP:
                        break;

                case FSMT_START:
                        _tr     = FSMT_NOP;
			_FSM = FSM_START;
                        break;

                case FSMT_STOP:
                        _tr     = FSMT_NOP;
			_FSM = FSM_STOP;
                        break;

	}

        //=========================
        // handle state

        switch (_FSM) {

                default:
			wait = 0;
                        break;

		case FSM_START:
			//
			//  --------------------------------- acquisisco dati attuali per successivo "restore" -----------------
			//
			//  se fallisce... si ferma senza ulteriori azioni
			//
		     {
			gettimeofday(&scan_start, NULL);
        		ss.str("");
        		ss << "FSM(START)";
        		prtverbose(1, obj, ss.str());

			AxisG2Scan_ns::scan_t axis_actual_scan_data = pdev->get_axis_actual_scan_data();
        		#ifdef _DBG_START
                		cout << "axis_actual_scan_data.begin          = " << axis_actual_scan_data.begin << endl;
                		cout << "axis_actual_scan_data.end            = " << axis_actual_scan_data.end << endl;
                		cout << "axis_actual_scan_data.sp_ac_dc.speed = " << axis_actual_scan_data.sp_ac_dc.speed << endl;
                		cout << "axis_actual_scan_data.sp_ac_dc.accel = " << axis_actual_scan_data.sp_ac_dc.accel << endl;
                		cout << "axis_actual_scan_data.sp_ac_dc.decel = " << axis_actual_scan_data.sp_ac_dc.decel << endl;
        		#endif /* _DBG_START */

        		string msg("FSM_START:: ");
        		try {
                		pdev->refresh_spacdc(msg);
                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_START:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_START:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}

			AxisG2Scan_ns::sp_ac_dc_t axis_default_data = pdev->get_axis_default_data();
        		#ifdef _DBG_START
                		cout << _name << " : axis_default_data.valid : " << (axis_default_data.valid?"VALID":"NOTvalid")  << endl;
                		cout << _name << " : axis_default_data.speed : " << axis_default_data.speed  << endl;
                		cout << _name << " : axis_default_data.accel : " << axis_default_data.accel  << endl;
                		cout << _name << " : axis_default_data.decel : " << axis_default_data.decel  << endl;
        		#endif /* _DBG_START */

			if (!axis_default_data.valid) {
				trace_err = timestamp() + "FSM_START:: axis_default_data.valid NOT valid";
                		cout << msg << endl;
				trace_err = timestamp() + msg;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
			}

		     	if ( _FSM == FSM_STOP ) break;

			_FSM = FSM_POSITIONING;
		     }
                        break;

		case FSM_POSITIONING:
			//
			//  --------------------------------- posizionamento sulla partenza... -----------------
			//
			//  .... per velocizzare: prima controllo se sono gia' sul punto di partenza
			//       se si, allora NON faccio un write_attribute(), che porta axisDevice
			//       a fare i suoi cicli...
			//
			//  se fallisce... si ferma senza ulteriori azioni
			//
		     {
        		ss.str("");
        		ss << "FSM(POSITIONING to begin)";
        		prtverbose(1, obj, ss.str());

			AxisG2Scan_ns::scan_t axis_actual_scan_data = pdev->get_axis_actual_scan_data();

        		string msg("FSM_POSITIONING:: ");
			Tango::DeviceAttribute beginvalue("Position", axis_actual_scan_data.begin);
        		try {
				double act_pos_double;
				Tango::DeviceAttribute act_pos = pdev->get_axis_dp()->read_attribute("Position");
				act_pos >> act_pos_double;

				if ( abs(act_pos_double - axis_actual_scan_data.begin) > 1.1 * pdev->get_axis_default_data().accuracy ){

					#ifdef _DBG_FSM_POSITIONING
						cout << "FSM_POSITIONING : going to position to begin...... " << endl;
					#endif /* _DBG_FSM_POSITIONING */

                			pdev->get_axis_dp()->write_attribute(beginvalue);

					//
					// ... il seguente usleep per dar modo di iniziare il moto,
					// ed in modo che il successivo stato dell FSM possa vedere un MOVING....
					// MA: forse e' meglio non fare questo usleep e fare che il successivo stato
					// della FSM controlli l'effettivo raggiungimento della posizione beginvalue.
					// 
					#ifdef _FSM_POSITIONING_WAIT_ON_STATE_
						usleep(500000);
					#endif
					//
					_FSM = FSM_POSITIONING_WAIT;
				} else {

					#ifdef _DBG_FSM_POSITIONING
						cout << "FSM_POSITIONING : ALREADY at begin position...... " << endl;
					#endif /* _DBG_FSM_POSITIONING */

					_FSM = FSM_SCAN;
				}

				pdev->set_state(Tango::MOVING);

                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_POSITIONING:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_POSITIONING:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
		     }
                        break;

		case FSM_POSITIONING_WAIT:
			//
			//  --------------------------------- attesa del posizionamento sulla partenza... -----------------
			//
			//  se fallisce... si ferma senza ulteriori azioni
			//
		     {
        		ss.str("");
        		ss << "FSM(POSITIONING_WAIT)";
        		prtverbose(1, obj, ss.str());

        		string msg("FSM_POSITIONING_WAIT:: ");
        		try {
				#ifdef _FSM_POSITIONING_WAIT_ON_STATE_
                			if ( pdev->get_axis_state() == Tango::MOVING ) break;
				#else
					// qui non ho ancora cambiato il tempo di polling e pertanto
					// posso chiedere ed OTTENERE il dato attuale di posizione.
					//
					AxisG2Scan_ns::scan_t axis_actual_scan_data = pdev->get_axis_actual_scan_data();
					Tango::DeviceAttribute act_pos = pdev->get_axis_dp()->read_attribute("Position");
					double pos_val;
					act_pos >> pos_val;

					#ifdef _DBG_FSM_POSITIONING_WAIT
						cout << "pos_val = " << pos_val << endl;
						cout << "pdev->get_axis_state() = " << Tango::DevStateName[pdev->get_axis_state()] << endl;
					#endif /* _DBG_FSM_POSITIONING_WAIT */

					if ( abs(pos_val - axis_actual_scan_data.begin) > 1.1 * pdev->get_axis_default_data().accuracy ) break;
                			if ( pdev->get_axis_state() == Tango::MOVING ) break;
				#endif
				_FSM = FSM_SCAN;
                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "POSITIONING_WAIT:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "POSITIONING_WAIT:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
		     }
                        break;

		case FSM_SCAN:
			//
			//  0: --------------------------------- imposto velocita' ed accelerazioni della scansione... -----------------
			//  1: --------------------------------- leggo il BN attuale... -----------------
			//  2: --------------------------------- POLL OFF... -----------------
			//  3: --------------------------------- scansione!... -----------------
			//
			//  se fallisce... si ferma senza ulteriori azioni ... passando per lo stato di FSM_STOP, si ha il ripristino
			//	ai valori precedenti
			//
		     {
        		ss.str("");
        		ss << "FSM(SCAN)";
        		prtverbose(1, obj, ss.str());

			AxisG2Scan_ns::scan_t axis_actual_scan_data = pdev->get_axis_actual_scan_data();

        		string msg("FSM_SCAN 0:: ");
			Tango::DeviceAttribute beginvalue("Position", axis_actual_scan_data.end);

			vector<Tango::DeviceAttribute> attr_in;
			string att1_name("Speed");
			string att2_name("Acceleration");
			string att3_name("Deceleration");

			attr_in.push_back(Tango::DeviceAttribute(att3_name, axis_actual_scan_data.sp_ac_dc.decel));
			attr_in.push_back(Tango::DeviceAttribute(att2_name, axis_actual_scan_data.sp_ac_dc.accel));
			attr_in.push_back(Tango::DeviceAttribute(att1_name, axis_actual_scan_data.sp_ac_dc.speed));
			#if 0
				cout << " -- going to write: " << endl;
				cout << "axis_actual_scan_data.sp_ac_dc.speed = " << axis_actual_scan_data.sp_ac_dc.speed << endl;
				cout << "axis_actual_scan_data.sp_ac_dc.accel = " << axis_actual_scan_data.sp_ac_dc.accel << endl;
				cout << "axis_actual_scan_data.sp_ac_dc.decel = " << axis_actual_scan_data.sp_ac_dc.decel << endl;
				cout << " -- -- -- -- -- -- " << endl;
			#endif

        		try {
                		pdev->get_axis_dp()->write_attributes(attr_in);
                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_SCAN 0:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_SCAN 0:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
		     	if ( _FSM == FSM_STOP ) break;

			//================================
        		msg = "FSM_SCAN 1:: ";
        		try {
                		Tango::DeviceData dd = pdev->get_axisBuffer_dp()->command_inout("GetLastBN");
				#ifdef _PPC_PPC_
					long lastbn;
				#else
					int lastbn;
				#endif
				dd >> lastbn;
                		pdev->set_attr_ScanBeginBN(lastbn);
        		} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_SCAN 1:: Tango Exception";
                		cout << msg << endl;
                		Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_SCAN 1:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
		     	if ( _FSM == FSM_STOP ) break;

#ifdef __FSM_SCAN_2__
			_FSM = FSM_SCAN_2;

        		ss.str("");
        		ss << "FSM(" << FSMName[_FSM] << ") FSM_SCAN to FSM_SCAN_2 [ " << FSMName[_FSM] << " ]";
        		prtverbose(2, obj, ss.str());
		     }
                        break;

		case FSM_SCAN_2:
			//
			//  2: --------------------------------- POLL OFF... -----------------
			//  3: --------------------------------- scansione!... -----------------
			//
			//  se fallisce... si ferma senza ulteriori azioni ... passando per lo stato di FSM_STOP, si ha il ripristino
			//	ai valori precedenti
			//
		     {
        		ss.str("");
        		ss << "FSM(SCAN_2)";
        		prtverbose(1, obj, ss.str());
#else
		     }
		     {

#endif
        		string msg("");
			//=============POLOFF===================
        		msg = "FSM_SCAN 2:: ";

			// Attenzione: la velocita' richiesta/impostata al passo precedente, puo' non essere quella
			// EFFETTIVAMENTE in atto (il Galil fa' delle 'approssimazioni")
			// Quindi, per calcolare il tempo e' opportuno rileggere la velocita'

			AxisG2Scan_ns::sp_ac_dc_t axis_actual_data = pdev->getspacdc(msg);;
			AxisG2Scan_ns::scan_t axis_actual_scan_data = pdev->get_axis_actual_scan_data();

			if (pdev->get_axisPollOff()){	// se la funzionalita' e' abilitata dalla relativa property allora POLL ON/OFF
				double lp = 10;  // in secondi
				int ScanFSMPollPeriod;
        			try {
					int long_poll = 10000;  // in ms
					if ( axis_actual_data.speed != 0.0 ) {
//
//   NOTA-01
//
					   #if 0
						lp = abs((axis_actual_scan_data.end-axis_actual_scan_data.begin)/axis_actual_data.speed);
						if ( axis_actual_data.accel != 0.0 ){
							lp += abs(axis_actual_data.speed/axis_actual_data.accel);
						} else {
							lp += .2;
						}
						if ( axis_actual_data.decel != 0.0 ){
							lp += abs(axis_actual_data.speed/axis_actual_data.decel);
						} else {
							lp += .2;
						}
					   #else /* 0 */

						lp = get_scan_calculated_time(
								axis_actual_scan_data.end-axis_actual_scan_data.begin,
								axis_actual_data.speed,
								axis_actual_data.accel,
								axis_actual_data.decel);

						if ( lp <= 0 ) lp = 5;
					   #endif /* 0 */

						// lp [s] .... long_poll [ms]
						long_poll = (int)(1000 * lp * 1.01); // era * 1.5 era 1.08
						if ( long_poll < MIN_LONG_POLL ) long_poll = MIN_LONG_POLL;

						#ifdef _DBG_POLLOFF
							cout << "lp = " << lp << " | long_poll = " << long_poll << endl;
						#endif /* _DBG_POLLOFF */

						// dal tempo di polling impostato per ls FSM e dal tempo di "long poll" che si impostera' del driver (YAMS/Galil)
						// calcolo il numero intero di cicli che la FSM dovra' aspettare prima di poter chiedere, con qualche possibilita'
						// di successo e senza timeout dei dati ai server che si appoggiano al driver, driver che dopo lo stato di "ibernazione"
						// durato "long poll" dovra' essere "rivitalizzato" al suo originario tempo di polling.
						//
						ScanFSMPollPeriod = pdev->get_ScanFSMPollPeriod();
						if ( ScanFSMPollPeriod != 0.0 ) {
							wait = ceil((float)long_poll / (float)ScanFSMPollPeriod) + 1; // era ... 1000 * lp ... ed anche + 1;
						} else {
							wait = 1;
						}
						long_poll = wait * ScanFSMPollPeriod;

					} else {
						// se axis_actual_scan_data.sp_ac_dc.speed === 0.0 allora non si deve fare nulla! ma terminare
                				cout << msg << "axis_actual_scan_data.sp_ac_dc.speed = " << axis_actual_scan_data.sp_ac_dc.speed << "!!!" << endl;
						cout << "---" << endl;
                        			_tr     = FSMT_NOP;
                        			_FSM  = FSM_STOP;
					}

					#ifdef _DBG_POLLOFF
						cout << "axis_actual_data.speed = " << axis_actual_data.speed << endl;
						cout << "lp = " << lp << " | long_poll = " << long_poll << " | wait = " << wait;
						cout << " i.e.: " << (wait*ScanFSMPollPeriod) << " ms"  << endl;
					#endif /* _DBG_POLLOFF */

					bool r = pdev->drv_attrpoll_off(long_poll);

        			} catch(Tango::DevFailed e) {
					trace_err = timestamp() + "FSM_SCAN 2:: Tango Exception";
                			cout << msg << endl;
                			Tango::Except::print_exception(e);
					cout << "---" << endl;
                        		_tr     = FSMT_NOP;
                        		_FSM  = FSM_STOP;
        			} catch (...) {
					trace_err = timestamp() + "FSM_SCAN 2:: General Exception";
                			cout << msg << endl;
					cout << "---" << endl;
                        		_tr     = FSMT_NOP;
                        		_FSM  = FSM_STOP;
        			}
			}
		     	if ( _FSM == FSM_STOP ) break;

			//=============SCAN!===================
			Tango::DeviceAttribute endvalue("Position", axis_actual_scan_data.end);
        		msg = "FSM_SCAN 3:: ";
        		try {
				#ifdef _DBG_FSM_POSITIONING_WAIT
					cout << "pdev->get_axis_state() = " << Tango::DevStateName[pdev->get_axis_state()] << endl;
				#endif /* _DBG_FSM_POSITIONING_WAIT */

                		pdev->get_axis_dp()->write_attribute(endvalue);
				pdev->set_state(Tango::MOVING);

                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_SCAN 3:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_SCAN 3:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
		     	if ( _FSM == FSM_STOP ) break;

			_FSM = FSM_SCAN_WAIT_TIME;

        		ss.str("");
        		ss << "FSM(" << FSMName[_FSM] << ") FSM_SCAN to FSM_SCAN_WAIT_TIME [ " << FSMName[_FSM] << " ]";
        		prtverbose(2, obj, ss.str());
		     }
                        break;

		case FSM_SCAN_WAIT_TIME:
			//
			//  --------------------------------- attesa del posizionamento sulla fine scansione... -----------------
			//
		     {
        		ss.str("");
        		ss << "FSM(SCAN_WAIT_TIME [" << FSMName[_FSM] << "])";
			ss << " wait = " << wait;
        		prtverbose(1, obj, ss.str());
        		string msg("FSM_SCAN_WAIT_TIME:: ");

			if ( wait > 0 ) wait--;
			else {
        			try {
					if ( abs(pdev->get_axis_position(FSM_SCAN_WAIT_TIME) - pdev->get_axis_actual_scan_data().end) < 1.1 * pdev->get_axis_default_data().accuracy ) _FSM = FSM_SCAN_END;
					else _FSM = FSM_SCAN_WAIT;

	                       	} catch(Tango::DevFailed e) {
					trace_err = timestamp() + "FSM_SCAN_WAIT_TIME:: Tango Exception";
	                		cout << msg << endl;
	                                Tango::Except::print_exception(e);
					cout << "---" << endl;
	                        	_tr     = FSMT_NOP;
	                        	_FSM  = FSM_STOP;
	        		} catch (...) {
					trace_err = timestamp() + "FSM_SCAN_WAIT_TIME:: General Exception";
	                		cout << msg << endl;
					cout << "---" << endl;
	                        	_tr     = FSMT_NOP;
	                        	_FSM  = FSM_STOP;
	        		}
			}
		      }
                        break;

		case FSM_SCAN_WAIT:
			//
			//  --------------------------------- attesa di fine movimento... -----------------
			//
		     {
        		ss.str("");
        		ss << "FSM(SCAN_WAIT)";
        		prtverbose(1, obj, ss.str());
        		string msg("FSM_SCAN_WAIT:: ");
//
//   NOTA-00
//
        		try {
				#if 0
                		if ( pdev->get_axis_state() == Tango::MOVING ) break; /// !!!!!!!!!!!!!!!!! NON posso fare cio': siamo ancora in LONG_POLL !!!!
				#endif

				if ( abs(pdev->get_axis_position(FSM_SCAN_WAIT) - pdev->get_axis_actual_scan_data().end) < 1.1 * pdev->get_axis_default_data().accuracy ) _FSM = FSM_SCAN_END;
				#ifdef _DBG_TRACE_FSM
				cout << __FILE__ << "@" << __LINE__ << " get_axis_position(FSM_SCAN_WAIT)=" << pdev->get_axis_position(FSM_SCAN_WAIT) << endl;
				#endif

                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_SCAN_WAIT:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_SCAN_WAIT:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}

		     }
                        break;

		case FSM_SCAN_END:
			//
			//  0: --------------------------------- leggo il BN attuale... -----------------
			//  1: --------------------------------- leggo i buffer circolari -----------------
			//
			//  se fallisce... si ferma senza ulteriori azioni ... passando per lo stato di FSM_STOP, si ha il ripristino
			//	ai valori precedenti
			//
		     {
        		ss.str("");
        		ss << "FSM(SCAN_END)";
        		prtverbose(1, obj, ss.str());

			//================================
        		string msg("FSM_SCAN_END 0:: ");
        		try {
                		Tango::DeviceData dd = pdev->get_axisBuffer_dp()->command_inout("GetLastBN");
				#ifdef _PPC_PPC_
					long lastbn;
				#else
					int lastbn;
				#endif
				dd >> lastbn;
                		pdev->set_attr_ScanEndBN(lastbn);
				#ifdef _DBG_TRACE_BNS
					cout << "lastbn = " << lastbn << endl;
				#endif /* _DBG_TRACE_BNS */
                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_SCAN_END 0:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_SCAN_END 0:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
		     	if ( _FSM == FSM_STOP ) break;

			//================================
        		msg = "FSM_SCAN_END 1:: ";
        		try {
				Tango::DeviceData dd;
				Tango::DevVarLongArray argin;
				argin.length(4);
				argin[0] = pdev->get_channel_no();
				argin[1] = 7;
				argin[2] = pdev->get_attr_ScanBeginBN();
				argin[3] = pdev->get_attr_ScanEndBN();
				#ifdef _DBG_TRACE_BNS
					cout << "command_inout(GetValues, dd)" << endl;
					cout << "argin[0] = " << argin[0] << endl;
					cout << "argin[1] = " << argin[1] << endl;
					cout << "argin[2] = " << argin[2] << endl;
					cout << "argin[3] = " << argin[3] << endl;
				#endif /* _DBG_TRACE_BNS */
				dd << argin;
                		dd = pdev->get_axisBuffer_dp()->command_inout("GetValues", dd);
				//cout << dd << endl;

				//pdev->set_attr_ScanPositions(dd);
				pdev->set_attr_ScanPositionsBN(dd);

				#ifdef _DBG_TRACE_BNS
				{
					//cout << "resp. dd.size() = " << endl;
				}
				#endif /* _DBG_TRACE_BNS */

        		} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_SCAN_END 1:: Tango Exception";
                		cout << msg << endl;
                		Tango::Except::print_exception(e);
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		} catch (...) {
				trace_err = timestamp() + "FSM_SCAN_END 1:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                        	_tr     = FSMT_NOP;
                        	_FSM  = FSM_STOP;
        		}
			//================================

                       	_FSM  = FSM_STOP;
		     }
                        break;

		case FSM_STOP:
			//
			//  0: --------------------------------- POLL ON... -----------------
			//
			//  se fallisce... riprova passando per FSM_STOP_ERROR
			//
		     {
        		ss.str("");
        		ss << "FSM(FSM_STOP)";
        		prtverbose(1, obj, ss.str());

        		string msg("FSM_STOP 0:: ");
			//=============POLOFF===================
			if (pdev->get_axisPollOff()){	// se la funzionalita' e' abilitata dalla relativa property allora POLL ON/OFF
//
//   NOTA-02
//
        			try {
					bool r = pdev->drv_attrpoll_on();
					if (pdev->get_axis_state() == Tango::MOVING) {
                        			_FSM  = FSM_STOP_WAIT;
					} else {
						do_stop_1();
		     				if ( _FSM == FSM_STOP_ERROR ) {
		      					pdev->set_state(Tango::ALARM);
						} else {  

                     					_FSM  = FSM_WAIT;
		      					pdev->set_state(Tango::ON);

							struct timeval t1, tr;
							gettimeofday(&t1, NULL);
							timersub(&t1, &scan_start, &tr);
							scantotaltime = (double)tr.tv_sec * 1000.0 + ((double)tr.tv_usec) / 1000.0;
						}
					}
                        		_tr     = FSMT_NOP;
        			} catch(Tango::DevFailed e) {
					trace_err = timestamp() + "FSM_STOP 0:: Tango Exception";
                			cout << msg << endl;
                			Tango::Except::print_exception(e);
					cout << "---" << endl;
                        		_tr     = FSMT_NOP;
                        		_FSM  = FSM_STOP_ERROR;
        			} catch (...) {
					trace_err = timestamp() + "FSM_STOP 0:: General Exception";
                			cout << msg << endl;
					cout << "---" << endl;
                        		_tr     = FSMT_NOP;
                        		_FSM  = FSM_STOP_ERROR;
        			}
			}
			//================================
		     }
		     	break;   

		case FSM_STOP_WAIT:
			//
			//  --------------------------------- attendiamo che axisDevice esca da MOVING... -----------------
			//
			//  se fallisce... riprova passando per FSM_STOP_ERROR
			//
		     {
        		ss.str("");
        		ss << "FSM(FSM_STOP_WAIT)";
        		prtverbose(1, obj, ss.str());

		   #if 0
			try {
				bool rr = pdev->refresh();
			} catch (...) {
				cout << "ERROR ob refresh() !" << endl;
			}
		   #endif

			// Bisogna dar modo al axisDevice di aggiornarsi
        		#ifdef _DBG_FSM_STOP_WAIT
				cout << timestamp() << "pdev->get_axis_state() = " << Tango::DevStateName[pdev->get_axis_state()] << endl;
        		#endif /* _DBG_FSM_STOP_WAIT */

        		string msg = "FSM_STOP WAIT:: ";
			try {
				while (pdev->get_axis_state() == Tango::MOVING) {
        				#if defined(_DBG_FSM_STOP_WAIT) && 0
						cout << timestamp() << " MOVING" << endl;
        				#endif /* _DBG_FSM_STOP_WAIT */
					//usleep(ceil((float)pdev->get_max_axisDevice_poll_time() * 1000.0 ));
				}
        			#ifdef _DBG_FSM_STOP_WAIT
					cout << timestamp() << "pdev->get_axis_state() = " << Tango::DevStateName[pdev->get_axis_state()] << endl;
        			#endif /* _DBG_FSM_STOP_WAIT */
                       		_tr     = FSMT_NOP;
                       		_FSM  = FSM_STOP_1;
                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_STOP WAIT:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                       		_tr     = FSMT_NOP;
                       		_FSM  = FSM_STOP_ERROR;
        		} catch (...) {
				trace_err = timestamp() + "FSM_STOP WAIT:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                       		_tr     = FSMT_NOP;
                       		_FSM  = FSM_STOP_ERROR;
        		}
		     }
		     	break;   


		case FSM_STOP_1:
			//
			//  1: --------------------------------- ripristino i valori -----------------
			//
			//  se fallisce... riprova passando per FSM_STOP_ERROR
			//
        		ss.str("");
        		ss << "FSM(FSM_STOP_1)";
        		prtverbose(1, obj, ss.str());

//AAA
			do_stop_1();
		     	if ( _FSM == FSM_STOP_ERROR ) {
		      		pdev->set_state(Tango::ALARM);
				break;
			}  

                     	_tr     = FSMT_NOP;
                     	_FSM  = FSM_WAIT;
		      	pdev->set_state(Tango::ON);

			struct timeval t1, tr;
			gettimeofday(&t1, NULL);
			timersub(&t1, &scan_start, &tr);
			scantotaltime = (double)tr.tv_sec * 1000.0 + ((double)tr.tv_usec) / 1000.0;

                        break;

		case FSM_STOP_ERROR:
			//
			//  0: --------------------------------- POLL ON... -----------------
			//  1: --------------------------------- ripristino i valori -----------------
			//
			//  se fallisce... RIMANE in FSM_STOP_ERROR ed in ALARM.
			//
		     {
        		ss.str("");
        		ss << "FSM(FSM_STOP_ERROR)";
        		prtverbose(1, obj, ss.str());

        		string msg("FSM_STOP_ERROR :: ");
			// try again pollon & restore ap ac dc
        		try {
				if (pdev->get_axisPollOff()){	// se la funzionalita' e' abilitata dalla relativa property allora POLL ON/OFF
					bool r = pdev->drv_attrpoll_on();
				}

				AxisG2Scan_ns::sp_ac_dc_t axis_default_data = pdev->get_axis_default_data();

				vector<Tango::DeviceAttribute> attr_in;
				string att1_name("Speed");
				string att2_name("Acceleration");
				string att3_name("Deceleration");

				attr_in.push_back(Tango::DeviceAttribute(att1_name, axis_default_data.speed));
				attr_in.push_back(Tango::DeviceAttribute(att2_name, axis_default_data.accel));
				attr_in.push_back(Tango::DeviceAttribute(att3_name, axis_default_data.decel));

                		pdev->get_axis_dp()->write_attributes(attr_in);

				struct timeval t1, tr;
				gettimeofday(&t1, NULL);
				timersub(&t1, &scan_start, &tr);
				scantotaltime = (double)tr.tv_sec * 1000.0 + ((double)tr.tv_usec) / 1000.0;

                       		_FSM  = FSM_WAIT;

                       	} catch(Tango::DevFailed e) {
				trace_err = timestamp() + "FSM_STOP_ERROR:: Tango Exception";
                		cout << msg << endl;
                                Tango::Except::print_exception(e);
				cout << "---" << endl;
                       		_tr     = FSMT_NOP;
                       		_FSM  = FSM_STOP_ERROR;
        		} catch (...) {
				trace_err = timestamp() + "FSM_STOP_ERROR:: General Exception";
                		cout << msg << endl;
				cout << "---" << endl;
                       		_tr     = FSMT_NOP;
                       		_FSM  = FSM_STOP_ERROR;
        		}
		     }

		     	if ( _FSM == FSM_STOP_ERROR ) break;

                       	_tr     = FSMT_NOP;
                       	_FSM  = FSM_WAIT;
			pdev->set_state(Tango::ON);

                        break;

	}

	return _FSM;
}

} /* namespace FSM_ns */
